#HINGO: same function as wsvm3 in matlab, which solves weighted SVM problem using Hinge loss as an approximation of 0-1 loss
# X: feature variables
# Y: Reward/outcome
# A: Assigned treatment
# PrTx: P(A=a|X=x)
# C: cost parameter/ tuning variable
# newX: feature variables for validation set/test set
library(kernlab)

#HINGE_SVM is an improved version of HINGEO, where we modify the bound of alpha as [0,C]
HINGE_SVM <- function(X, A, Y, PrTx, C, newX ){
  # process the data
  W  <- lm(Y ~ X)$residuals
  Wabs <- abs(W)/PrTx #weight
  Z <- (W>0) - (W<=0)
  As <- Z * A #Label
  # modified Q 
  Q <- X%*%t(X) * ((As*Wabs)%*%t(As*Wabs)) 
  n <- length(Y)
  #solve alpha1=alpha/Wabs
  #fit <- ipop(-Wabs, Q, A=t(As*Wabs), b=0, l=rep(0, n), u=rep(C,n), r=0 )
  #check if ipop returns an error
  fit <- tryCatch(ipop(-Wabs, Q, A=t(As*Wabs), b=0, l=rep(0, n), u=rep(C,n), r=0 ), error=function(e) e)
  stop_loop <- 0
  max.error <- 0
  #check if fit returns error due to sigularity of Q
  while ("error" %in% class(fit)){
    if (max.error < 5){
      print(fit)
      Q = Q + diag(n) * 0.0001
      fit <- tryCatch(ipop(-Wabs, Q, A=t(As*Wabs), b=0, l=rep(0, n), u=rep(C,n), r=0 ), error=function(e) e)
      max.error <- max.error + 1
    }else {
      stop_loop <- 1; break;
    }
  }
  if (stop_loop == 0){
    alpha1 <- primal(fit)
    alpha <- alpha1 * Wabs
  
    beta <- t(X) %*% (As* alpha)
    #index that subject to 0<=\alpha_i<=C|W_i|/pi
    index = (alpha1 >C * 1e-4 )& (alpha1 < C*(1-1e-4))
    beta0_pred <-As - X%*% beta
    
    if (sum(index) > 0){
      beta0 <- mean(beta0_pred[index])
    } else {
    #By Andrew Ng's paper
      beta0 <- -max((As==-1)*beta0_pred)/2 - min((As==1)*beta0_pred)/2
    }
  }else{
    beta0 <- sign (mean (Y*(A==1)/PrTx) - mean (Y*(A==-1)/PrTx))
    beta <- rep(0, ncol(X))
  }
     
  ITR <- (cbind(rep(1, nrow(newX)), newX)%*% c(beta0,beta)>0)*2-1
  #print(table(ITR))
  return(list( ITR=ITR, e.beta0=beta0,e.beta=beta));
  # for linear kernel t(X)%*%e.beta are the coefficients;  
}

#predit ITR based on rule fit
HINGE_SVM.pred <- function(fit, newX){
  ITR <-  (cbind(rep(1, nrow(newX)), newX)%*% c(fit$e.beta0,fit$e.beta)>0)*2-1
  return(ITR)
}


# Finding the best tuning parameters C using K-fold cross-validation
# bigC: candidate values for C
# K: number of folds
HINGE.CV  <- function(X, A, Y, PrTx, bigC, K){  
  #sample K folds
  folds <- sample(1:K, length(Y), replace=T)
  Vtemp <- -99999
  OC <- bigC[1]
  for(i in 1:length(bigC)){
    predV <- rep(0, K)
    for (k in 1:K){
      Ytrain <- Y[folds!=k]; Xtrain <- X[folds!=k,]; Atrain <- A[folds!=k]; PTtrain <- PrTx[folds!=k];
      Ytest <- Y[folds==k]; Xtest <- X[folds==k,];  Atest <- A[folds==k]; PTtest <- PrTx[folds==k];
      #estimate rule based on traning set and benefit on test set
      temp <- HINGE_SVM(Xtrain, Atrain, Ytrain, PTtrain, bigC[i], Xtest)
      predV[k] <- mean(Ytest*(Atest==temp$ITR)/PTtest)
    }
    if (mean(predV)>Vtemp){
      OC <- bigC[i]
      Vtemp <- mean(predV)
    }
  }
  return(list(optC=OC,Value=Vtemp)); 
}





