library(kernlab)
ramp.loss <- function(u,s){
  (u/s <= -1/2) + (u/s > -1/2) * (u/s < 1/2) * (-u/s + 1/2)
}
#RAMPO3: same function as ramp_svm2 in matlab, which solves weighted SVM problem using RAMP loss as an approximation of 0-1 loss
# X: feature variables (should be standardized)
# Y: Reward/outcome
# A: Assigned treatment
# PrTx: P(A=a|X=x)
# C: cost parameter/ tuning variable
# s: tuning parameter of ramp loss
# iter.Max: maximum number of interation for step2 in the algorithm
# tol: tolerance for convergence
# newX: feature variables for validation set/test set
# beta.init: initial value of estimates, if not pre-specified, will use the results estimated by Hinge_svm
# RAMP_SVM is an improved version of RAMPO3, where we modify the bound of alpha as [0,C]
RAMP_SVM <- function(X, A, Y, PrTx, C, s, newX, iter.Max, tol, beta.init=NULL ){
  # process the data
  W  <- lm(Y ~ X)$residuals
  Wabs <- abs(W)/PrTx
  Z <- (W>0) - (W<=0)
  As <- Z*A
  Q <- X%*%t(X) * (As%*%t(As))
  Q1 <- X%*%t(X) * ((As*Wabs)%*%t(As*Wabs)) 
  n <- length(Y)
  p <- dim(X)[2]
  delta <- 99999
  
  #if no initial value is set, set it by HINGE_svm estimator
  if(is.null(beta.init)){
    fit_temp <- HINGE_SVM(X, A, Y, PrTx, C, newX)
    beta.init <- c(fit_temp$e.beta0, fit_temp$e.beta)
    # As X is standardized, so the chance of all X1...Xp >100 is almost 0
    # If beta are too small comparing to beta0, then beta0 will dominate the decision rule
  }
  #beta.init <-c(0, rep(5, p))
  eta.init <- As * cbind(rep(1, n), X)%*% beta.init
  ramploss.init <- C * sum(Wabs*ramp.loss(eta.init, s)) + 0.5 * sum (beta.init[-1]^2)
  stop_loop <- 0
  iter <- 1
  #iteratively solve qudartive programming unitl ramp.loss converges
  while (iter < iter.Max ){
    dh2 <- -(eta.init < -0.5 * s)/s
    f <- -(1 - 2*C*Q %*% (Wabs* dh2)/s)/2 *Wabs
   
    fit <- tryCatch(ipop(f, Q1/s^2, A=t(As)/s*Wabs, b=-C*sum(As*Wabs*dh2), l=rep(0, n), u=rep(C,n), r=0 ), error=function(e) e)
    #check if fit returns error due to sigularity of Q
    max.error <- 0
    #break the loop when this while loop cannot stop and return beta with the most updated beta
    while ("error" %in% class(fit)){
      if (max.error < 5){
        # print(fit)
        Q <- Q + diag(n) * 0.0001
        fit <- tryCatch(ipop(f, Q1/s^2, A=t(As)/s*Wabs, b=-C*sum(As*Wabs*dh2), l=rep(0, n), u=rep(C,n), r=0 ), error=function(e) e)
        max.error <- max.error + 1
      }else{
        stop_loop <- 1
        break
      }
    }
  # fit <- ipop(f, Q1/s^2, A=t(As)/s*Wabs, b=-C*sum(As*Wabs*dh2), l=rep(0, n), u=rep(C,n), r=0 )
    if (stop_loop ==1) {
      break
    }
   #alpha1 = alpha /Wabs
    alpha1 <- primal(fit)
    alpha <- alpha1 * Wabs
    
    beta.pred <- t(X) %*% (As* (C*Wabs*dh2 + alpha/s))
    # beta.pred <-  t(X) %*% (As* C*Wabs*dh2/PrTx + alpha/s)
    #index that subject to 0<=\alpha_i<=C|W_i|/pi
    #index = (alpha >1e-4 )& (alpha <C*Wabs-1e-4)
    beta0 <- s/As/2 - X%*% beta.pred
    index <- (alpha1 >C * 1e-4 )& (alpha1 < C*(1-1e-4))
    if (sum(index) > 0){
      beta0.pred <- mean(beta0[index])
    } else {
      #AA = (alpha1 > C*(1-1e-4) ) & (A == As) | (alpha1 < C * 1e-4 ) & (A == -As)       
      #BB = (alpha1 > C*(1-1e-4) ) & (A == -As) | (alpha1 < C * 1e-4 ) & (A == As) 
      #beta0.pred <- (min(beta0[AA==1]) + max(beta0[BB==1]))/2
      
      #By Andrew Ng's paper
      beta0.pred <- -max((As==-1)*beta0)/2 - min((As==1)*beta0)/2
    }
 
   
    #beta0.pred <- mean(s/As/2 - X%*% beta.pred)
    #beta0 <-X%*% beta.pred
    #beta0.pred <- -max((As==-1)*beta0)/2 - min((As==-1)*beta0)/2
    beta.new <- c(beta0.pred, beta.pred)
    
    eta.new <- As * cbind(rep(1, n), X)%*% beta.new
    ramploss.new <-  C * sum(Wabs*ramp.loss(eta.new, s)) + 0.5 * sum (beta.new[-1]^2)
    
    delta <- abs(ramploss.new-ramploss.init)
    #print(beta.init)
    #print(ramploss.init)
    if (delta < tol){
      break
    }
    iter <- iter + 1
    beta.init <- beta.new
    eta.init <- eta.new
    ramploss.init <- ramploss.new    
  }
  
  ITR <- (cbind(rep(1, nrow(newX)), newX)%*% beta.init>0)*2-1
  #print(table(ITR))
  #return(list(delta=delta, ITR=ITR, e.beta0=beta0.pred,e.beta=beta.pred,ramploss = ramploss.init))
  return(list(delta=delta, ITR=ITR, e.beta0=beta.init[1],e.beta=beta.init[-1],ramploss = ramploss.init))
  # for linear kernel t(X)%*%e.beta are the coefficients;  
}

#predit ITR based on rule fit
RAMP_SVM.pred <- function(fit, newX){
  ITR <-  (cbind(rep(1, nrow(newX)), newX)%*% c(fit$e.beta0,fit$e.beta)>0)*2-1
  return(ITR)
}

# Finding the best tuning parameters C and s using K-fold cross-validation
# bigC: candidate values for C
# bigS: candidate values for S
# K: number of folds
RAMP.CV  <- function(X, A, Y, PrTx, bigC, bigS, K, iter.Max, tol, beta.init=NULL){  
  #sample K folds
  folds <- sample(1:K, length(Y), replace=T)
  Vtemp <- -99999
  OC <- bigC[1]; OS <- bigS[1]
  for(i in 1:length(bigC)){
    #if beta.init is not set, estimate based on full sample an given C to avoid one  estimation for each C, S and at different folds of training data.
    if(is.null(beta.init)){
      fit_temp <- HINGE_SVM(X, A, Y, PrTx, bigC[i], X )
      beta.init <- c(fit_temp$e.beta0, fit_temp$e.beta)
    }
    for (j in 1:length(bigS)){      
      predV <- rep(0, K)
      for (k in 1:K){
        Ytrain <- Y[folds!=k]; Xtrain <- X[folds!=k,]; Atrain <- A[folds!=k]; PTtrain <- PrTx[folds!=k];
        Ytest <- Y[folds==k]; Xtest <- X[folds==k,];  Atest <- A[folds==k]; PTtest <- PrTx[folds==k];
        #estimate rule based on traning set and benefit on test set
        temp <- RAMP_SVM(Xtrain, Atrain, Ytrain, PTtrain, bigC[i], bigS[j], Xtest, iter.Max, tol, beta.init)
        predV[k] <- mean(Ytest*(Atest==temp$ITR)/PTtest)
      }
      if (mean(predV)>Vtemp){
        OC <- bigC[i]; OS <- bigS[j]
        Vtemp <- mean(predV)
      }
      #print(c("predv",predV))
    }
  }
  return(list(optC=OC, optS=OS, Value=Vtemp)); 
}
